require 'singleton'

module Bydo
  class TypeMatch
    include Singleton
    
    def ok?
      true
    end
  end
  
  class TypeMismatch
    attr_reader :expected, :actual

    def ok?
      false
    end
    
    def initialize(expected, actual)
      @expected = expected
      @actual = actual
    end
  end
  
  class ModuleChecker
    def initialize(mod)
      @mod = mod
    end

    def type_str
      @mod.to_s
    end

    def check(obj)
      if obj.kind_of?(@mod) || obj.nil?
        TypeMatch.instance
      else
        TypeMismatch.new(type_str, obj.class.to_s)
      end
    end
  end
  
  class ArrayChecker
    def initialize(elem_types)
      @elem_types = elem_types
    end

    def type_str
      "[" + @elem_types.map{|e| e.type_str}.join(",") + "]"
    end
    
    def check(obj)
      return TypeMatch.instance if obj.nil?
      
      unless obj.kind_of?(Array) 
        return TypeMismatch.new(type_str, obj.class.to_s)
      end

      obj.each do |elem|
        unless @elem_types.any?{|t| t.check(elem).ok?}
          return TypeMismatch.new(type_str, "[#{elem.class.to_s}]")
        end
      end

      TypeMatch.instance
    end
  end

  class NotNilChecker
    def initialize(type)
      @type = type
    end

    def check(obj)
      if obj.nil?
        TypeMismatch.new("non-nil", "nil")
      else
        @type.check(obj)
      end
    end
  end

  class BoolChecker
    include Singleton
    
    def check(obj)
      if obj == true || obj == false
        TypeMatch.instance
      else
        TypeMismatch.new("boolean", obj.class.to_s)
      end
    end
  end
  
  class NotNil
    attr_reader :t
    
    def initialize(t)
      @t = t
    end
  end

  class Bool
    include Singleton
  end
  
  class Checker
    def initialize(types, rest, ret)
      @types = types.map{|t| create_checker(t)}
      @rest = rest ? ArrayChecker.new(create_checker(rest)) : nil
      @ret = ret ? create_checker(ret) : ModuleChecker.new(Object)
    end

    def create_checker(typeobj)
      case typeobj
      when Module
        ModuleChecker.new(typeobj)
      when Array
        ArrayChecker.new(typeobj.map{|elem| create_checker(elem)})
      when NotNil
        NotNilChecker.new(typeobj.t)
      when Bool
        BoolChecker.instance
      else
        raise TypeError, "bad type definition"
      end
    end

    def nth_str(num)
      case num
      when 1 then "1st"
      when 2 then "2nd"
      when 3 then "3rd"
      else "#{num}th"
      end
    end

    def check_arg(t, arg, msg, bt)
      result = t.check(arg)
      unless result.ok?
        raise(TypeError,
              "#{msg} should be #{result.expected}, " +
              "but #{result.actual}",
              caller(bt))
      end
    end
    
    def check(args)
      @types.zip(args).each_with_index do |(t,arg),idx|
        check_arg(t, arg, "#{nth_str(idx+1)} argument", 7)
      end

      if @rest
        check_arg(@rest, args.last(args.size-@types.size), "rest arguments", 6)
      end
    end

    def check_retval(val)
      check_arg(@ret, val, "return value", 5)
    end

    def check_arity(arity)
      if (arity >= 0 && @types.size != arity) ||
          (arity >= 0 && @rest ) ||
          (arity < 0 && @types.size > arity) ||
          (arity < 0 && !@rest)
        raise TypeError, "type declaration is bad"
      end
    end
  end

  def _rtyper_ckeckers
    @_rtyper_checkers
  end

  def notnil(type)
    NotNil.new(type)
  end

  def Bool()
    Bool.instance
  end
  
  def declare(method, *types)
    if types.last.kind_of?(Hash)
      hash = types.pop
    else
      hash = {}
    end
    
    @_rtyper_checkers ||= Hash.new
    @_rtyper_in_def = false
    @_rtyper_checkers[method] = Checker.new(types, hash[:*], hash[:ret])
  end
  
  def method_added(name)
    return unless @_rtyper_checkers && @_rtyper_checkers[name] && !@_rtyper_in_def
    @_rtyper_checkers[name].check_arity(instance_method(name).arity)
      
    impl_name = "_rtyper_impl_#{name.id2name}"
    alias_method(impl_name, name)
    
    @_rtyper_in_def = true
    module_eval <<-EOS
      def #{name}(*args,&block)
        checker = self.class._rtyper_ckeckers[:#{name}]
        checker.check(args)
        retval = #{impl_name}(*args, &block)
        checker.check_retval(retval)
        retval
      end
    EOS
    @_rtyper_in_def = false
  end
end

if __FILE__ == $0
  def trap_exc
    begin
      yield
    rescue TypeError => e
      puts "#{e.backtrace.first}: #{e.message}"
    end
  end


  class Heke
    extend Bydo
    
    declare :f, Integer, :ret => Integer
    def f(n)
      n * 2
    end
    
    declare :g, Integer
    def g(n)
    end
    
    declare :h, Integer, :ret => Integer
    def h(n)
      "a"
    end
    
    declare :k, [Numeric], :ret => Numeric
    def k(nums)
      nums.inject{|x,y| x+y}
    end

    declare :l, Integer, String, :ret => String
    def l(n, str)
      str * n
    end

    declare :m, Integer, :ret => String
    def m(n)
      nil
    end

    declare :n, :ret => [String]
    def n
      nil
    end

    declare :o, notnil(String)
    def o(s)
    end

    declare :q, Bool()
    def q(b)
      b
    end

  end

  class Hoge
    extend Bydo
    
    declare :f, Float, :ret => Float
    def f(x)
      x * 2.0
    end
  end
  
  p Heke.new.f(3)
  p Hoge.new.f(3.0)
  p Heke.new.f(3)
  p Hoge.new.f(3.0)
  trap_exc{ p Heke.new.f("abc") }
  p Heke.new.g(4)
  trap_exc{ p Heke.new.g(3.1) }
  trap_exc{ p Heke.new.h(3) }
  p Heke.new.k((1..10).to_a)
  p Heke.new.k([1.2, 2.3, 2])
  trap_exc{ p Heke.new.k([1.2, 2.3, 2, "a"])}
  p Heke.new.l(3, "abc")
  trap_exc{ p Heke.new.l("abc", 3)}
  trap_exc{ p Heke.new.l(3, 3)}
  p Heke.new.m(4)
  p Heke.new.n
  trap_exc{ p Heke.new.o }
  p Heke.new.q(true)
  p Heke.new.q(false)
  trap_exc{ p Heke.new.q(nil) }
end
